/************************************************
* DBASE.C - dBASE file I/O routines				*
*												*
* 900514 MCMason - adapted to MSC v5.1			*
* 8911?? BRagsdell - originally written TPascal	*
************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <string.h>
#include "dbase.h"

#define Day$upd	update[2]	/*Last update: day	*/
#define Mon$upd	update[1]	/*Last update: month*/
#define Yr$upd	update[0]	/*Last update: year	*/

#define F(x) (*x->fld)

/************************************************
* dbfOpen - open a dBASE III+/IV data file.		*
*												*
* INP:	fname - file name						*
* OUT:	dbfFILE pointer, or NULL if error.		*
************************************************/
dbfFILE dbfOpen(const char *fname)
	{
	int i;
	FILE *f;
	dbfFILE r;
	struct _dbfHdrMask
		{
		char		Version;
		char		update[3];
		long		NumRex;
		int			HdrLen;
		int			RecLen;
		char		Reserved[20];
		} h;
	struct _dbfFldMask
		{
		char		Name[11];
		char		Type;
		char		Reserved1[4];
		unsigned char Length;
		unsigned char Dec;
		char		Reserved2[14];
		} FldTmp;

	/* Allocate dBASE file data block */
	r=(dbfFILE)malloc(sizeof(struct _dbfFILE));

	/* Open the file */
	f = fopen(fname,"rb");
	if (f==NULL)
		goto dbfOpenErrorExit;
	r->file = f;

	/* Read the dBASE header data */
	if (fread(&h,sizeof(h),1,r->file) != 1)
		goto dbfOpenErrorExit2;
	r->Version = h.Version & 0x07;
	r->Day$upd = h.Day$upd;
	r->Mon$upd = h.Mon$upd;
	r->Yr$upd  = h.Yr$upd;
	r->NumRex  = h.NumRex;
	r->HdrLen  = h.HdrLen;
	r->RecLen  = h.RecLen;
	r->NumFlds = (h.HdrLen - 33)/32;
	r->FileSize= h.HdrLen+h.RecLen*h.NumRex+1;
	if (r->Version != 3)
		goto dbfOpenErrorExit2;

	/* Reallocate the RAM block, now that we	*/
	/* know how many fields to get ...			*/
	r->fld = malloc(sizeof(struct _dbfFLD)
									*r->NumFlds);
	if (r->fld == NULL)
		goto dbfOpenErrorExit2;

	/* Read in all the field descriptors */
	for (i=0; i<r->NumFlds; i++)
		{
		if (fread(&FldTmp,sizeof(FldTmp),1,
									r->file)!=1)
			goto dbfOpenErrorExit2;
		strncpy(F(r)[i].Name,FldTmp.Name,11);
		F(r)[i].Type   = FldTmp.Type;
		F(r)[i].Length = FldTmp.Length;
		F(r)[i].Dec    = FldTmp.Dec;
		}

	/* Done ... return pointer to data */
	return r;

	/*******************************************
	* Only get here when an error has occurred *
	*******************************************/
dbfOpenErrorExit2:		/*---When file is open--*/
	fclose(f);
dbfOpenErrorExit:		/*--Before file is open-*/
	free((void *)r);
	return NULL;		/* Return error code	*/
	}

/************************************************
* dbfClose - close a dBASE file					*
*												*
* INP:	dbf - dbase file handle to close		*
* OUT:	TRUE if error, FALSE otherwise			*
************************************************/
int dbfClose(dbfFILE dbf)
	{
	if (fclose(dbf->file))
		return TRUE;
	free(dbf->fld);
	free(dbf);
	return FALSE;
	}


/************************************************
* dbfDspHdr - display Dbase file header info	*
*												*
* INP:	hdr - pointer to file header structure	*
************************************************/
static char buf[50];
void dbfDspHdr(dbfFILE hdr, FILE *out)
	{
	fprintf(out,"dBASE Version         :%8.8s\n",
			itoa(hdr->Version,buf,10));
	fprintf(out,"Number of data records:%8.8s\n",
			ltoa(hdr->NumRex,buf,10));
	fprintf(out,"Date of last update   : %2.2s/"
			"%s/%s\n",itoa(hdr->Mon$upd,buf,10),
			itoa(hdr->Day$upd,buf+10,10),
			itoa(hdr->Yr$upd, buf+20,10));
	fprintf(out,"Header length         :%8.8s\n",
			itoa(hdr->HdrLen,buf,10));
	fprintf(out,"Record length         :%8.8s\n",
			itoa(hdr->RecLen,buf,10));
	fprintf(out,"Number of fields      :%8.8s\n",
			itoa(hdr->NumFlds,buf,10));
	fprintf(out,"File size             :%8.8s\n",
			ltoa(hdr->FileSize,buf,10));
	}


/************************************************
* dbfDspStru - display the structure of a dBASE	*
* file:  Name, Field type, length and number of	*
* decimals (if numeric).						*
************************************************/
void dbfDispStru(dbfFILE hdr, FILE *out)
	{
	char	*Ty;
	int		i;

	fputs("\nField  Field Name  Type        Width"
		 "     Dec\n",out);
	
	for (i=0; i<hdr->NumFlds; i++)
		{
		switch (F(hdr)[i].Type)
			{
			case 'C': Ty = "Character  "; break;
			case 'L': Ty = "Logical    "; break;
			case 'N': Ty = "Number     "; break;
			case 'F': Ty = "Floating Pt"; break;
			case 'D': Ty = "Date       "; break;
			case 'M': Ty = "Memo       "; break;
			default:  Ty = "Unknown    "; break;
			}
		fprintf(out,"%5d  %-11.11s%11.11s  %3d"
			"      %2d\n",i+1,F(hdr)[i].Name,Ty,
			F(hdr)[i].Length,F(hdr)[i].Dec);
		}
	itoa(hdr->RecLen,buf,10);
	fprintf(out,"   ** Total **                %4s\n",
		buf);
	}

/************************************************
* dbfReadRec - read dBASE record to buffer		*
************************************************/
int dbfReadRec(dbfFILE hdr, long Rec, char *b)
	{
	long CurrentPos;

	CurrentPos=(Rec-1)*hdr->RecLen+hdr->HdrLen;
	fseek(hdr->file,CurrentPos,SEEK_SET);
	if (fread(b,hdr->RecLen,1,hdr->file) != 1)
		return TRUE;
	return FALSE;
	}

/************************************************
* dbfPrintRec - print the record in the buffer	*
************************************************/
void dbfPrintRec(dbfFILE hdr, FILE *out, long Rec,
		char *b)
	{
	int FPos, i;

	fprintf(out,"%3d  ",Rec);

	/* Print Active/Deleted marker */
	fputc(*b,out);
	fputc(' ',out);

	/* Print the fields */
	FPos=1;
	for (i=0; i<hdr->NumFlds; i++)
		{
		switch (F(hdr)[i].Type)
			{
			case 'D':
				fprintf(out,"%2.2s/%2.2s/%4.4s",
					b+FPos+4,b+FPos+6,b+FPos);
				break;
			case 'L':
				switch (b[FPos])
					{
					case 'Y':
					case 'T':
						fputs(".T.",out);
						break;
					case 'N':
					case 'F':
						fputs(".F.",out);
						break;
					default:
						fputs(".?.",out);
					}
				break;
			case 'M':	/* Ignore MEMO fields */
				break;
			default:
				sprintf(buf,"%%%d.%ds",
						F(hdr)[i].Length,
						F(hdr)[i].Length);
				fprintf(out,buf,b+FPos);
			}
		fputc(' ',out);
		FPos += F(hdr)[i].Length;
		}
	fputc('\n',out);
	}


/************************************************
* dbfList - Dbase file printing routine.  This	*
* routine prints a header, & each record in file*
************************************************/
void dbfList(dbfFILE d, FILE *out)
	{
	long i;
	char *buf;

	if ( (buf=malloc(d->RecLen+2)) == NULL )
		return;
	fputc('\n',stdout);
	for (i=1; i<=d->NumRex; i++)
		{
		dbfReadRec(d, i, buf);
		dbfPrintRec(d, out, i, buf);
		}
	free(buf);
	}
