/************************************************************
* DBC.C - added routines to create dBase IV files			*
************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dos.h>
#include "dbase.h"

#define Day$upd update[2]	// Day of last update
#define Mon$upd update[1]	// Month of last update
#define Yr$upd  update[0]	// Year of last update
#define F(x) (*x->fld)		// Field access

/************************************************************
* dbfCreate - create a dBase file from the description in	*
* the specified header.										*
*															*
* INP:	hdr - a standard dBase file header					*
*		name - name of the file to create					*
* OUT:	TRUE on error										*
************************************************************/
int dbfCreate(dbfFILE hdr, char *name)
	{
	FILE *f;
	int   i;
	struct _dbfHdrMask
		{
		char	Version;
		char	update[3];
		long	NumRex;
		int		HdrLen;
		int		RecLen;
		char	Reserved[20];
		} h;
	struct _dbfFldMask
		{
		char	Name[11];
		char	Type;
		char	Reserved1[4];
		unsigned char Length;
		unsigned char Dec;
		char	Reserved2[14];
		} FldTmp;

	/* Open the specified file */
	f = fopen(name,"wb");
	if (f==NULL)
		return TRUE;

	/* Write the dBASE header data */
	h.Version = (char)(hdr->Version & 0x07);
	h.Day$upd = hdr->Day$upd;
	h.Mon$upd = hdr->Mon$upd;
	h.Yr$upd  = hdr->Yr$upd;
	h.NumRex  = 0L;
	h.HdrLen  = hdr->HdrLen;
	h.RecLen  = hdr->RecLen;
	memset(h.Reserved,0,20);
	if (fwrite(&h,sizeof(h),1,f) != 1)
		{
		fclose(f);
		return TRUE;
		}

	/* Write out the list of fields */
	memset(FldTmp.Reserved1,0,4);
	memset(FldTmp.Reserved2,0,14);
	for (i=0; i<hdr->NumFlds; i++)
		{
		memset(FldTmp.Name,0,11);
		strncpy(FldTmp.Name,F(hdr)[i].Name,11);
		FldTmp.Type   = F(hdr)[i].Type;
		FldTmp.Length = (unsigned char)F(hdr)[i].Length;
		FldTmp.Dec    = (unsigned char)F(hdr)[i].Dec;
		if (fwrite(&FldTmp,sizeof(FldTmp),1,f) != 1)
			{
			fclose(f);
			return TRUE;
			}
		}

	/* Write the header termination character & quit */
	fputc(0x0d,f);
	fclose(f);
	return FALSE;
	}

/****************************************************************
* dbfInputStr - read the database structure from an input file	*
*																*
* INP:	name - name of the input file containing dBase struct.	*
* OUT:	NULL if error, otherwise, a dbfFILE described by file.	*
****************************************************************/
dbfFILE dbfInputStr(char *name)
	{
	FILE   *inFile;
	char    workbuf[61];
	int     cur, max;
	dbfFILE r;
	struct  dosdate_t date;

	inFile = fopen(name,"r");
	if (!inFile) return NULL;

	// First, search for the word 'Field'
	//
	while (fgets(workbuf,60,inFile)!=NULL)
		if (!strncmp(workbuf,"Field",5))
			break;
	if (feof(inFile))
		{
		fclose(inFile);
		return NULL;
		}

	// Create a dbfFILE header
	//
	r=(dbfFILE)malloc(sizeof(dbfFILE));
	r->Version = 3;
	_dos_getdate(&date);
	r->Day$upd = date.day;
	r->Mon$upd = date.month;
	r->Yr$upd  = (char)(date.year%100);
	r->NumRex  = 0L;
	r->HdrLen  = 0;
	r->RecLen  = 0;
	r->NumFlds = 0;
	r->FileSize= 0L;
	r->file    = NULL;
	r->fld     = (dbfFLD *)NULL;

	// Next, read in the field descriptions
	//
	cur=max=r->RecLen=0;
	while (fgets(workbuf,60,inFile)!=NULL)
		{
		if (!strncmp(workbuf,"   ** Total **",14))
			break;
		if (cur >= max)
			{
			max += 5;
			r->fld=(dbfFLD (*)[])realloc(r->fld,max*sizeof(dbfFLD));
			if (!r->fld)
				{
				fclose(inFile);
				free(r);
				return NULL;
				}
			}
		sscanf(workbuf,"%*d %s %c%*10c %d %d",F(r)[cur].Name,
			&F(r)[cur].Type,&F(r)[cur].Length,&F(r)[cur].Dec);
		r->RecLen += F(r)[cur].Length;
		++cur;
		}

	// Adjust heading fields
	//
	r->RecLen++;			// +1 for CR (at end of record)
	r->NumFlds = cur;
	r->HdrLen  = (cur+1)*32 + 1;
	r->FileSize= r->HdrLen;
	fclose(inFile);
	return r;
	}
