// mtcopy.c RHS 5/25/90
/****************************************************************
* MTCOPY.C - multithread programming with file copy demo		*
****************************************************************/

#define INCL_DOS
#define INCL_ERRORS
#include<os2.h>
#include<stdio.h>
#include<stdlib.h>
#include<stdarg.h>
#include<string.h>
#include<process.h>

#define MAXSEGSIZE      65535u
#define THREADSTACKSIZE 1024

typedef struct segment
    {
    ULONG   segSem;     // semaphore
    SEL     segSel;     // selector
    USHORT  segfull;    // empty/full flag
    USHORT  segamount;  // # of bytes read
    } SEG;
typedef SEG *PSEG;

#define MAXSEGS 7

typedef struct copyparams
    {
    ULONG   copySem;
    char    *copysource;
    char    *copydest;
    USHORT  copystatus;
    HFILE   copyinput;
    HFILE   copyoutput;
    SEG     seg[MAXSEGS];
    } COPY;
typedef COPY *PCOPY;


void read_thread(void far *);
void write_thread(void far *);
void error_exit(USHORT errcode, char *fmt,...);
void mtcopy(COPY *copy);
void mtcopy_init(PCOPY *cp, char *source, char *dest);
USHORT mtcopy_finished(COPY *copy);

void read_thread(void far *c)
    {
    USHORT  end_thread = FALSE;
    USHORT  rtc = 0;
    PCOPY   copy = (COPY *)c;
    PSEG    seg = &copy->seg[0];

    while(!end_thread)
        {
        DosSemRequest(&(seg[rtc].segSem),SEM_INDEFINITE_WAIT);
        if(!seg[rtc].segfull)
            {
            USHORT errcode;

            printf("reading %s into seg[%u]...",
            			copy->copysource,rtc);

            if(errcode = DosRead(copy->copyinput,
                    MAKEP(seg[rtc].segSel,0),MAXSEGSIZE,
                    &(seg[rtc].segamount)))
                     error_exit(errcode,"Error %u reading %s\n",
                                errcode,copy->copysource);
            if(!seg[rtc].segamount)  		// nothing read:
                {                       	// :End of file
                DosClose(copy->copyinput);	// close input
                end_thread = TRUE;      	// terminate thread
                }
            else
                printf("read %u bytes into seg[%u] from %s.\n",
                    seg[rtc].segamount, rtc, copy->copysource);
            seg[rtc].segfull = TRUE;    // for write & main
            }                           // threads
        DosSemClear(&(seg[rtc].segSem));
        if(++rtc == MAXSEGS)
            rtc = 0;
        }
    printf("finished reading %s, terminating\n",
    		copy->copysource);
    _endthread();
    }

void write_thread(void far *c)
    {
    USHORT  end_thread = FALSE;
    USHORT  wtc = 0;
    COPY    *copy = (COPY *)c;
    SEG     *seg = &copy->seg[0];

    while(!end_thread)
        {
        DosSemRequest(&seg[wtc].segSem,SEM_INDEFINITE_WAIT);
        if(seg[wtc].segfull)
            {
            USHORT errcode;

            if(!seg[wtc].segamount)     	// EndOfFile reached
                {
                DosClose(copy->copyoutput); // close output
                end_thread = TRUE;      	// terminate thread
                DosSemClear(&copy->copySem);
                }
            else
                {
                printf("writing %s from seg[%u]...\n",
                		copy->copydest,wtc);

                if(errcode = DosWrite(copy->copyoutput,
                         MAKEP(seg[wtc].segSel,0),
                         seg[wtc].segamount,&seg[wtc].segamount))
                    error_exit(errcode,"Error %u writing %s\n",
                               errcode,copy->copydest);
                seg[wtc].segfull = FALSE;
                }
            }
        DosSemClear(&seg[wtc].segSem);
        if(++wtc == MAXSEGS)
            wtc = 0;
        }
    printf("finished writing %s, terminating successfully\n",
        copy->copydest);
    _endthread();
    }


void error_exit(USHORT errcode, char *fmt,...)
    {
    va_list argptr;

    va_start(argptr,fmt);
    vprintf(fmt,argptr);
    va_end(argptr);

    DosExit(errcode,EXIT_PROCESS);
    }


void mtcopy(COPY *copy)
    {
    USHORT  errcode,action;
    SEG     *seg = &copy->seg[0];

    DosSemRequest(&copy->copySem,SEM_INDEFINITE_WAIT);

    if (errcode=DosOpen(copy->copysource,&copy->copyinput,
    		&action,0L,0,FILE_OPEN,OPEN_ACCESS_READONLY |
            OPEN_SHARE_DENYWRITE,0L))
        error_exit(errcode,"Error %u opening %s\n",
                   errcode,copy->copysource);

    if(errcode = DosOpen(copy->copydest,&copy->copyoutput,
    		&action,0L,FILE_NORMAL,FILE_CREATE|FILE_TRUNCATE,
            OPEN_ACCESS_WRITEONLY | OPEN_SHARE_DENYWRITE, 0L))
        error_exit(errcode,"Error %u opening/creating %s\n",
                   errcode,copy->copydest);

    {
    int i;
    for( i = 0; i < MAXSEGS; i++)
        if(errcode = DosAllocSeg(MAXSEGSIZE,&seg[i].segSel,
                SEG_NONSHARED))
            error_exit(errcode,"Error %u allocating"
                    " buffer\n",errcode);
    }

    if(_beginthread(read_thread,NULL,
            THREADSTACKSIZE,(void far *)copy) == -1)
        error_exit(0,"ERROR: start read thread\n");
    if(_beginthread(write_thread,NULL,
            THREADSTACKSIZE,(void far *)copy) == -1)
        error_exit(0,"ERROR: start write thread\n");
    }

USHORT mtcopy_finished(COPY *copy)
    {
    USHORT ret;

    if(!(ret=DosSemRequest(&copy->copySem,SEM_IMMEDIATE_RETURN)))
        return TRUE;
    return FALSE;
    }

void mtcopy_init(PCOPY *cp, char *source, char *dest)
    {
    PCOPY p;

    p = (PCOPY)malloc(sizeof(COPY));
    memset(p,0,sizeof(COPY));	// initialize copy parameters
    p->copysource = source;
    p->copydest = dest;
    *cp = p;
    }

void main(int argc, char **argv)
    {
    void *cp;

    if(argc != 3)
        error_exit(0,"Usage: MTCOPY <sourcefile>"
                        " <destinationfile>\n");

    mtcopy_init(&cp,strupr(argv[1]),strupr(argv[2]));
    mtcopy(cp);                  // begin multithread copy

    // substitute some other part of an application
    // here.  Copy is done by other threads

    while(TRUE)
        {
        if(mtcopy_finished(cp))
            {
            printf("main thread: copy complete, terminating\n");
            break;
            }
        }
    DosExit(0,EXIT_PROCESS);
    }
